@@ -18,7 +18,7 @@
 -- 
 -- * 'OverlappingGCDMonoid'
 
-{-# LANGUAGE Haskell2010, FlexibleInstances, Trustworthy #-}
+{-# LANGUAGE CPP, Haskell2010, FlexibleInstances, Trustworthy #-}
 
 module Data.Monoid.GCD (
    GCDMonoid(..),
@@ -33,10 +33,15 @@ import qualified Data.ByteString as ByteString
 import qualified Data.ByteString.Unsafe as ByteString
 import qualified Data.ByteString.Lazy as LazyByteString
 import qualified Data.Text as Text
+import qualified Data.Text.Encoding as TextEncoding
 import qualified Data.Text.Internal as Internal
 import qualified Data.Text.Internal.Lazy as LazyInternal
-import           Data.Text.Unsafe (lengthWord16, reverseIter)
+import           Data.Text.Unsafe (reverseIter)
+#if MIN_VERSION_text(2,0,0)
+import           Data.Text.Unsafe (Iter(..))
+#endif
 import qualified Data.Text.Lazy as LazyText
+import qualified Data.Text.Lazy.Encoding as LazyEncoding
 import qualified Data.IntMap as IntMap
 import qualified Data.IntSet as IntSet
 import qualified Data.Map as Map
@@ -383,15 +388,27 @@ instance LeftGCDMonoid Text.Text where
    stripCommonPrefix x y = maybe (Text.empty, x, y) id (Text.commonPrefixes x y)
 
 -- | @since 1.0
--- /O(suffixLength)/
+-- /O(suffixLength)/, except on GHCjs where it is /O(m+n)/
 instance RightGCDMonoid Text.Text where
-   stripCommonSuffix x@(Internal.Text xarr xoff xlen) y@(Internal.Text yarr yoff ylen) = go (pred xlen) (pred ylen)
+#if !ghcjs_HOST_OS
+  stripCommonSuffix x@(Internal.Text xarr xoff xlen) y@(Internal.Text yarr yoff ylen) = go (pred xlen) (pred ylen)
       where go i j | i >= 0 && j >= 0 && xc == yc = go (i+xd) (j+yd)
                    | otherwise = (Internal.text xarr xoff (succ i),
                                   Internal.text yarr yoff (succ j),
                                   Internal.text xarr (xoff+i+1) (xlen-i-1))
+#if MIN_VERSION_text(2,0,0)
+               where Iter xc xd = reverseIter x i
+                     Iter yc yd = reverseIter y j
+#else
                where (xc, xd) = reverseIter x i
                      (yc, yd) = reverseIter y j
+#endif
+#else
+  stripCommonSuffix x y =
+    let (xlist, ylist, slist) =
+          stripCommonSuffix (TextEncoding.encodeUtf8 x) (TextEncoding.encodeUtf8 y)
+    in (TextEncoding.decodeUtf8 xlist, TextEncoding.decodeUtf8 ylist, TextEncoding.decodeUtf8 slist)
+#endif
 
 -- Lazy Text instances
 
@@ -402,6 +419,7 @@ instance LeftGCDMonoid LazyText.Text where
 -- | @since 1.0
 -- /O(m+n)/
 instance RightGCDMonoid LazyText.Text where
+#if !ghcjs_HOST_OS
    stripCommonSuffix x0 y0
       | x0len < y0len = go id y0p id x0 y0s
       | x0len > y0len = go x0p id id x0s y0
@@ -411,7 +429,7 @@ instance RightGCDMonoid LazyText.Text where
             x0len = lazyLengthWord16 x0
             y0len = lazyLengthWord16 y0
             lazyLengthWord16 = LazyText.foldlChunks addLength 0
-            addLength n x = n + lengthWord16 x
+            addLength n x = n + (\(Internal.Text _ _ l) -> l) x
             splitWord16 xp 0 x = (xp, x)
             splitWord16 xp n (LazyInternal.Chunk x@(Internal.Text arr off len) xs)
                | n < len = (xp . LazyInternal.chunk (Internal.Text arr off n),
@@ -431,3 +449,9 @@ instance RightGCDMonoid LazyText.Text where
                | (x1p, y1p, c1s) <- stripCommonSuffix x y =
                     go (xp . cs . LazyInternal.chunk x1p) (yp . cs . LazyInternal.chunk y1p) (LazyInternal.chunk c1s) xs ys
             go _ _ _ _ _ = error "impossible"
+#else
+  stripCommonSuffix x y =
+    let (xlist, ylist, slist) =
+          stripCommonSuffix (LazyEncoding.encodeUtf8 x) (LazyEncoding.encodeUtf8 y)
+    in (LazyEncoding.decodeUtf8 xlist, LazyEncoding.decodeUtf8 ylist, LazyEncoding.decodeUtf8 slist)
+#endif
