@@ -1,5 +1,5 @@
 {-
-    Copyright 2014-2019 Mario Blazevic
+    Copyright 2014-2021 Mario Blazevic
 
     License: BSD3 (see BSD3-LICENSE.txt file)
 -}
@@ -10,15 +10,15 @@
 -- base monoid of 'LinePositioned' must be a 'TextualMonoid', but for the price it will keep track of the current line
 -- and column numbers as well.
 --
--- All positions are zero-based:
+-- Line number is zero-based, column one-based:
 --
 -- >> let p = pure "abcd\nefgh\nijkl\nmnop\n" :: LinePositioned String
 -- >> p
--- >Line 0, column 0: "abcd\nefgh\nijkl\nmnop\n"
+-- >"abcd\nefgh\nijkl\nmnop\n"
 -- >> Data.Monoid.Factorial.drop 13 p
--- >Line 2, column 3: "l\nmnop\n"
+-- >Line 2, column 4: "l\nmnop\n"
 
-{-# LANGUAGE Haskell2010 #-}
+{-# LANGUAGE Haskell2010, DeriveDataTypeable #-}
 
 module Data.Monoid.Instances.Positioned (
    OffsetPositioned, LinePositioned, extract, position, line, column
@@ -29,6 +29,7 @@ import Control.Applicative -- (Applicative(..))
 import qualified Data.List as List
 import Data.String (IsString(..))
 
+import Data.Data (Data, Typeable)
 import Data.Semigroup (Semigroup(..))
 import Data.Monoid (Monoid(..), Endo(..))
 import Data.Semigroup.Cancellative (LeftReductive(..), RightReductive(..))
@@ -56,7 +57,7 @@ data LinePositioned m = LinePositioned{fullOffset :: !Int,
                                        -- | the current line
                                        line :: !Int,
                                        lineStart :: !Int,
-                                       extractLines :: m}
+                                       extractLines :: m} deriving (Data, Typeable)
 
 -- | the current column
 column :: LinePositioned m -> Int
@@ -73,7 +74,7 @@ instance Applicative OffsetPositioned where
    OffsetPositioned _ f <*> OffsetPositioned p c = OffsetPositioned p (f c)
 
 instance Applicative LinePositioned where
-   pure = LinePositioned 0 0 0
+   pure = LinePositioned 0 0 (-1)
    LinePositioned _ _ _ f <*> LinePositioned p l lp c = LinePositioned p l lp (f c)
 
 instance Positioned OffsetPositioned where
@@ -97,9 +98,11 @@ instance Ord m => Ord (LinePositioned m) where
    compare LinePositioned{extractLines= a} LinePositioned{extractLines= b} = compare a b
 
 instance Show m => Show (OffsetPositioned m) where
+   showsPrec prec (OffsetPositioned 0 c) = showsPrec prec c
    showsPrec prec (OffsetPositioned pos c) = shows pos . (": " ++) . showsPrec prec c
 
 instance Show m => Show (LinePositioned m) where
+   showsPrec prec (LinePositioned 0 0 (-1) c) = showsPrec prec c
    showsPrec prec (LinePositioned pos l lpos c) =
       ("Line " ++) . shows l . (", column " ++) . shows (pos - lpos) . (": " ++) . showsPrec prec c
 
@@ -120,11 +123,9 @@ instance (StableFactorial m, TextualMonoid m) => Semigroup (LinePositioned m) wh
      | otherwise = LinePositioned p2' l2' lp2' c
      where c = mappend c1 c2
            p2' = max 0 $ p2 - length c1
-           lp2' = min p2' lp2
-           l2' = if l2 == 0 then 0 else max 0 $ l2 - Textual.foldl_' countLines 0 c1
-           countLines :: Int -> Char -> Int
-           countLines n '\n' = succ n
-           countLines n _ = n
+           lp2' = p2' - (p2 - lp2 - cd + 1)
+           l2' = if l2 == 0 then 0 else max 0 (l2 - ld)
+           (ld, cd) = linesColumns' c1
    {-# INLINE (<>) #-}
 
 instance (StableFactorial m, TextualMonoid m) => Monoid (LinePositioned m) where
@@ -242,7 +243,7 @@ instance StableFactorial m => Factorial (OffsetPositioned m) where
 
 instance (StableFactorial m, FactorialMonoid m) => FactorialMonoid (OffsetPositioned m) where
    splitPrimePrefix (OffsetPositioned p c) = fmap rewrap (splitPrimePrefix c)
-      where rewrap (cp, cs) = (OffsetPositioned p cp, OffsetPositioned (succ p) cs)
+      where rewrap (cp, cs) = (OffsetPositioned p cp, OffsetPositioned (if null cs then 0 else succ p) cs)
    splitPrimeSuffix (OffsetPositioned p c) = fmap rewrap (splitPrimeSuffix c)
       where rewrap (cp, cs) = (OffsetPositioned p cp, OffsetPositioned (p + length cp) cs)
    spanMaybe s0 f (OffsetPositioned p0 t) = rewrap $ Factorial.spanMaybe (s0, p0) f' t
@@ -275,28 +276,53 @@ instance (StableFactorial m, FactorialMonoid m) => FactorialMonoid (OffsetPositi
 
 instance (StableFactorial m, TextualMonoid m) => Factorial (LinePositioned m) where
    factors (LinePositioned p0 l0 lp0 c) = snd $ List.mapAccumL next (p0, l0, lp0) (factors c)
-      where next (p, l, lp) c1 | characterPrefix c1 == Just '\n' = ((succ p, succ l, p), LinePositioned p l lp c1)
-                               | otherwise = ((succ p, l, lp), LinePositioned p l lp c1)
+      where next (p, l, lp) c1 = let p' = succ p
+                                 in p' `seq` case characterPrefix c1
+                                             of Just '\n' -> ((p', succ l, p), LinePositioned p l lp c1)
+                                                Just '\f' -> ((p', succ l, p), LinePositioned p l lp c1)
+                                                Just '\r' -> ((p', l, p), LinePositioned p l lp c1)
+                                                Just '\t' -> ((p', l, lp + (p - lp) `mod` 8 - 8),
+                                                              LinePositioned p l lp c1)
+                                                Just ch | isZeroWidth ch -> ((p, l, lp), LinePositioned p l lp c1)
+                                                _ -> ((p', l, lp), LinePositioned p l lp c1)
    primePrefix (LinePositioned p l lp c) = LinePositioned p l lp (primePrefix c)
    foldl f a0 (LinePositioned p0 l0 lp0 c0) = fstOf4 $! Factorial.foldl f' (a0, p0, l0, lp0) c0
-      where f' (a, p, l, lp) c | characterPrefix c == Just '\n' = (f a (LinePositioned p l lp c), succ p, succ l, p)
-                               | otherwise = (f a (LinePositioned p l lp c), succ p, l, lp)
+      where f' (a, p, l, lp) c = case characterPrefix c
+                                 of Just '\n' -> (f a (LinePositioned p l lp c), succ p, succ l, p)
+                                    Just '\f' -> (f a (LinePositioned p l lp c), succ p, succ l, p)
+                                    Just '\r' -> (f a (LinePositioned p l lp c), succ p, l, p)
+                                    Just '\t' -> (f a (LinePositioned p l lp c), succ p, l, lp + (p - lp) `mod` 8 - 8)
+                                    Just ch | isZeroWidth ch -> (f a (LinePositioned p l lp c), p, l, lp)
+                                    _ -> (f a (LinePositioned p l lp c), succ p, l, lp)
    foldl' f a0 (LinePositioned p0 l0 lp0 c0) = fstOf4 $! Factorial.foldl' f' (a0, p0, l0, lp0) c0
       where f' (a, p, l, lp) c = let a' = f a (LinePositioned p l lp c)
-                                 in seq a' (if characterPrefix c == Just '\n'
-                                            then (a', succ p, succ l, p)
-                                            else (a', succ p, l, lp))
+                                 in seq a' (case characterPrefix c
+                                            of Just '\n' -> (a', succ p, succ l, p)
+                                               Just '\f' -> (a', succ p, succ l, p)
+                                               Just '\r' -> (a', succ p, l, p)
+                                               Just '\t' -> (a', succ p, l, lp + (p - lp) `mod` 8 - 8)
+                                               Just ch | isZeroWidth ch -> (a', p, l, lp)
+                                               _ -> (a', succ p, l, lp))
    foldr f a0 (LinePositioned p0 l0 lp0 c0) = Factorial.foldr f' (const3 a0) c0 p0 l0 lp0
-      where f' c cont p l lp
-               | characterPrefix c == Just '\n' = f (LinePositioned p l lp c) $ ((cont $! succ p) $! succ l) p
-               | otherwise = f (LinePositioned p l lp c) $ (cont $! succ p) l lp
+      where f' c cont p l lp = case characterPrefix c
+                               of Just '\n' -> f (LinePositioned p l lp c) $ ((cont $! succ p) $! succ l) p
+                                  Just '\f' -> f (LinePositioned p l lp c) $ ((cont $! succ p) $! succ l) p
+                                  Just '\r' -> f (LinePositioned p l lp c) $ (cont $! succ p) l p
+                                  Just '\t' -> f (LinePositioned p l lp c) $ (cont $! succ p) l
+                                               $! lp + (p - lp) `mod` 8 - 8
+                                  Just ch | isZeroWidth ch -> f (LinePositioned p l lp c) $ (cont p) l lp
+                                  _ -> f (LinePositioned p l lp c) $ (cont $! succ p) l lp
    foldMap f (LinePositioned p0 l0 lp0 c) = appEndo (Factorial.foldMap f' c) (const mempty) p0 l0 lp0
       where -- f' :: m -> Endo (Int -> Int -> Int -> m)
             f' prime = Endo (\cont p l lp-> f (LinePositioned p l lp prime)
-                                            <>
-                                            if characterPrefix prime == Just '\n'
-                                            then cont (succ p) (succ l) p
-                                            else cont (succ p) l lp)
+                                            `mappend`
+                                            case characterPrefix prime
+                                            of Just '\n' -> cont (succ p) (succ l) p
+                                               Just '\f' -> cont (succ p) (succ l) p
+                                               Just '\r' -> cont (succ p) l p
+                                               Just '\t' -> cont (succ p) l (lp + (p - lp) `mod` 8 - 8)
+                                               Just ch | isZeroWidth ch -> cont p l lp
+                                               _ -> cont (succ p) l lp)
    length = length . extractLines
    reverse (LinePositioned p l lp c) = LinePositioned p l lp (Factorial.reverse c)
    {-# INLINE primePrefix #-}
@@ -310,9 +336,15 @@ instance (StableFactorial m, TextualMonoid m) => Factorial (LinePositioned m) wh
 instance (StableFactorial m, TextualMonoid m) => FactorialMonoid (LinePositioned m) where
    splitPrimePrefix (LinePositioned p l lp c) = fmap rewrap (splitPrimePrefix c)
       where rewrap (cp, cs) = (LinePositioned p l lp cp,
-                               if characterPrefix cp == Just '\n'
-                               then LinePositioned (succ p) (succ l) p cs
-                               else LinePositioned (succ p) l lp cs)
+                               if null cs then mempty
+                               else case characterPrefix cp
+                                    of Just '\n' -> LinePositioned p' (succ l) p cs
+                                       Just '\f' -> LinePositioned p' (succ l) p cs
+                                       Just '\r' -> LinePositioned p' l p cs
+                                       Just '\t' -> LinePositioned p' l (lp + (p - lp) `mod` 8 - 8) cs
+                                       Just ch | isZeroWidth ch -> LinePositioned p l lp cs
+                                       _ -> LinePositioned p' l lp cs)
+            p' = succ p
    splitPrimeSuffix (LinePositioned p l lp c) = fmap rewrap (splitPrimeSuffix c)
       where rewrap (cp, cs) = (LinePositioned p l lp cp, LinePositioned p' (l + lines) (p' - columns) cs)
                where len = length cp
@@ -322,26 +354,38 @@ instance (StableFactorial m, TextualMonoid m) => FactorialMonoid (LinePositioned
       where f' (s, p, l, lp) prime = do s' <- f s (LinePositioned p l lp prime)
                                         let p' = succ p
                                             l' = succ l
-                                        Just $! p' `seq` if characterPrefix prime == Just '\n'
-                                                         then l' `seq` (s', p', l', p)
-                                                         else (s', p', l, lp)
+                                        Just $! p' `seq` case characterPrefix prime
+                                                         of Just '\n' -> l' `seq` (s', p', l', p)
+                                                            Just '\f' -> l' `seq` (s', p', l', p)
+                                                            Just '\r' -> (s', p', l, p)
+                                                            Just '\t' -> (s', p', l, lp + (p - lp) `mod` 8 - 8)
+                                                            Just ch | isZeroWidth ch -> (s', p, l, lp)
+                                                            _ -> (s', p', l, lp)
             rewrap (prefix, suffix, (s, p, l, lp)) = (LinePositioned p0 l0 lp0 prefix, LinePositioned p l lp suffix, s)
    spanMaybe' s0 f (LinePositioned p0 l0 lp0 c) = rewrap $! Factorial.spanMaybe' (s0, p0, l0, lp0) f' c
       where f' (s, p, l, lp) prime = do s' <- f s (LinePositioned p l lp prime)
                                         let p' = succ p
                                             l' = succ l
-                                        Just $! s' `seq` p' `seq` if characterPrefix prime == Just '\n'
-                                                                  then l' `seq` (s', p', l', p)
-                                                                  else (s', p', l, lp)
+                                        Just $! s' `seq` p' `seq` case characterPrefix prime
+                                                                  of Just '\n' -> l' `seq` (s', p', l', p)
+                                                                     Just '\f' -> l' `seq` (s', p', l', p)
+                                                                     Just '\r' -> (s', p', l, p)
+                                                                     Just '\t' -> (s', p', l, lp + (p - lp) `mod` 8 - 8)
+                                                                     Just ch | isZeroWidth ch -> (s', p, l, lp)
+                                                                     _ -> (s', p', l, lp)
             rewrap (prefix, suffix, (s, p, l, lp)) = (LinePositioned p0 l0 lp0 prefix, LinePositioned p l lp suffix, s)
 
    span f (LinePositioned p0 l0 lp0 t) = rewrap $ Factorial.spanMaybe' (p0, l0, lp0) f' t
       where f' (p, l, lp) prime = if f (LinePositioned p l lp prime)
                                   then let p' = succ p
                                            l' = succ l
-                                       in Just $! p' `seq` if characterPrefix prime == Just '\n'
-                                                           then l' `seq` (p', l', p)
-                                                           else (p', l, lp)
+                                       in Just $! p' `seq` case characterPrefix prime
+                                                           of Just '\n' -> l' `seq` (p', l', p)
+                                                              Just '\f' -> l' `seq` (p', l', p)
+                                                              Just '\r' -> (p', l, p)
+                                                              Just '\t' -> (p', l, lp + (p - lp) `mod` 8 - 8)
+                                                              Just c | isZeroWidth c -> (p, l, lp)
+                                                              _ -> (p', l, lp)
                                   else Nothing
             rewrap (prefix, suffix, (p, l, lp)) = (LinePositioned p0 l0 lp0 prefix, LinePositioned p l lp suffix)
    splitAt n m@(LinePositioned p l lp c) | n <= 0 = (mempty, m)
@@ -369,7 +413,8 @@ instance IsString m => IsString (LinePositioned m) where
    fromString = pure . fromString
 
 instance (StableFactorial m, TextualMonoid m) => TextualMonoid (OffsetPositioned m) where
-   splitCharacterPrefix (OffsetPositioned p c) = fmap (fmap $ OffsetPositioned $ succ p) (splitCharacterPrefix c)
+   splitCharacterPrefix (OffsetPositioned p t) = fmap rewrap (splitCharacterPrefix t)
+      where rewrap (c, cs) = if null cs then (c, mempty) else (c, OffsetPositioned (succ p) cs)
 
    fromText = pure . fromText
    singleton = pure . singleton
@@ -448,6 +493,8 @@ instance (StableFactorial m, TextualMonoid m) => TextualMonoid (OffsetPositioned
       where rewrap (prefix, suffix) = (OffsetPositioned p0 prefix, OffsetPositioned (p0 + length prefix) suffix)
    dropWhile_ bt fc t = snd (span_ bt fc t)
    takeWhile_ bt fc (OffsetPositioned p t) = OffsetPositioned p (takeWhile_ bt fc t)
+   toString ft (OffsetPositioned _ t) = toString (ft . pure) t
+   toText ft (OffsetPositioned _ t) = toText (ft . pure) t
 
    {-# INLINE characterPrefix #-}
    {-# INLINE splitCharacterPrefix #-}
@@ -470,11 +517,18 @@ instance (StableFactorial m, TextualMonoid m) => TextualMonoid (OffsetPositioned
    {-# INLINE find #-}
 
 instance (StableFactorial m, TextualMonoid m) => TextualMonoid (LinePositioned m) where
-   splitCharacterPrefix (LinePositioned p l lp c) =
-      case splitCharacterPrefix c
+   splitCharacterPrefix (LinePositioned p l lp t) =
+      case splitCharacterPrefix t
       of Nothing -> Nothing
-         Just ('\n', rest) -> Just ('\n', LinePositioned (succ p) (succ l) p rest)
-         Just (ch, rest) -> Just (ch, LinePositioned (succ p) l lp rest)
+         Just (c, rest) | null rest -> Just (c, mempty)
+         Just ('\n', rest) -> Just ('\n', LinePositioned p' (succ l) p rest)
+         Just ('\f', rest) -> Just ('\f', LinePositioned p' (succ l) p rest)
+         Just ('\r', rest) -> Just ('\r', LinePositioned p' l p rest)
+         Just ('\t', rest) -> Just ('\t', LinePositioned p' l (lp + (p - lp) `mod` 8 - 8) rest)
+         Just (ch, rest)
+            | isZeroWidth ch -> Just (ch, LinePositioned p l lp rest)
+            | otherwise -> Just (ch, LinePositioned p' l lp rest)
+      where p' = succ p
 
    fromText = pure . fromText
    singleton = pure . singleton
@@ -489,7 +543,12 @@ instance (StableFactorial m, TextualMonoid m) => TextualMonoid (LinePositioned m
    foldl ft fc a0 (LinePositioned p0 l0 lp0 c0) = fstOf4 $ Textual.foldl ft' fc' (a0, p0, l0, lp0) c0
       where ft' (a, p, l, lp) c = (ft a (LinePositioned p l lp c), succ p, l, lp)
             fc' (a, p, l, _lp) '\n' = (fc a '\n', succ p, succ l, p)
-            fc' (a, p, l, lp) c = (fc a c, succ p, l, lp)
+            fc' (a, p, l, _lp) '\f' = (fc a '\f', succ p, succ l, p)
+            fc' (a, p, l, _lp) '\r' = (fc a '\r', succ p, l, p)
+            fc' (a, p, l, lp) '\t' = (fc a '\t', succ p, l, lp + (p - lp) `mod` 8 - 8)
+            fc' (a, p, l, lp) c
+               | isZeroWidth c = (fc a c, p, l, lp)
+               | otherwise = (fc a c, succ p, l, lp)
    foldl' ft fc a0 (LinePositioned p0 l0 lp0 c0) = fstOf4 $ Textual.foldl' ft' fc' (a0, p0, l0, lp0) c0
       where ft' (a, p, l, lp) c = let a' = ft a (LinePositioned p l lp c)
                                       p' = succ p
@@ -497,21 +556,34 @@ instance (StableFactorial m, TextualMonoid m) => TextualMonoid (LinePositioned m
             fc' (a, p, l, lp) c = let a' = fc a c
                                       p' = succ p
                                       l' = succ l
-                                  in a' `seq` p' `seq` if c == '\n'
-                                                       then l' `seq` (a', p', l', p)
-                                                       else (a', p', l, lp)
+                                  in a' `seq` p' `seq` case c
+                                                       of '\n' -> l' `seq` (a', p', l', p)
+                                                          '\f' -> l' `seq` (a', p', l', p)
+                                                          '\r' -> (a', p', l, p)
+                                                          '\t' -> (a', p', l, lp + (p - lp) `mod` 8 - 8)
+                                                          _ | isZeroWidth c -> (a', p, l, lp)
+                                                          _ -> (a', p', l, lp)
    foldr ft fc a0 (LinePositioned p0 l0 lp0 c0) = Textual.foldr ft' fc' (const3 a0) c0 p0 l0 lp0
       where ft' c cont p l lp = ft (LinePositioned p l lp c) $ (cont $! succ p) l lp
             fc' c cont p l lp
                | c == '\n' = fc c $ ((cont $! succ p) $! succ l) p
+               | c == '\f' = fc c $ ((cont $! succ p) $! succ l) p
+               | c == '\r' = fc c $ (cont $! succ p) l p
+               | c == '\t' = fc c $ (cont $! succ p) l (lp + (p - lp) `mod` 8 - 8)
+               | isZeroWidth c = fc c $ (cont p) l lp
                | otherwise = fc c $ (cont $! succ p) l lp
 
    spanMaybe s0 ft fc (LinePositioned p0 l0 lp0 t) = rewrap $ Textual.spanMaybe (s0, p0, l0, lp0) ft' fc' t
       where ft' (s, p, l, lp) prime = do s' <- ft s (LinePositioned p l lp prime)
                                          let p' = succ p
                                          Just $! seq p' (s', p', l, lp)
-            fc' (s, p, l, lp) c = fc s c
-                                  >>= \s'-> Just $! seq p' (if c == '\n' then seq l' (s', p', l', p) else (s', p', l, lp))
+            fc' (s, p, l, lp) c =
+               fc s c
+               >>= \s'-> Just $! seq p' (if c == '\n' || c == '\f' then seq l' (s', p', l', p)
+                                         else if c == '\r' then (s', p', l, p)
+                                         else if c == '\t' then (s', p', l, lp + (p - lp) `mod` 8 - 8)
+                                         else if isZeroWidth c then (s', p, l, lp)
+                                         else (s', p', l, lp))
                where p' = succ p
                      l' = succ l
             rewrap (prefix, suffix, (s, p, l, lp)) = (LinePositioned p0 l0 lp0 prefix, LinePositioned p l lp suffix, s)
@@ -522,14 +594,24 @@ instance (StableFactorial m, TextualMonoid m) => TextualMonoid (LinePositioned m
             fc' (s, p, l, lp) c = do s' <- fc s c
                                      let p' = succ p
                                          l' = succ l
-                                     Just $! s' `seq` p' `seq` (if c == '\n' then seq l' (s', p', l', p) else (s', p', l, lp))
+                                     Just $! s' `seq` p'
+                                             `seq` (if c == '\n' || c == '\f' then seq l' (s', p', l', p)
+                                                    else if c == '\r' then (s', p', l, p)
+                                                    else if c == '\t' then (s', p', l, lp + (p - lp) `mod` 8 - 8)
+                                                    else if isZeroWidth c then (s', p, l, lp)
+                                                    else (s', p', l, lp))
             rewrap (prefix, suffix, (s, p, l, lp)) = (LinePositioned p0 l0 lp0 prefix, LinePositioned p l lp suffix, s)
    span ft fc (LinePositioned p0 l0 lp0 t) = rewrap $ Textual.spanMaybe' (p0, l0, lp0) ft' fc' t
       where ft' (p, l, lp) prime = if ft (LinePositioned p l lp prime)
                                    then let p' = succ p
                                         in p' `seq` Just (p', l, lp)
                                    else Nothing
-            fc' (p, l, lp) c | fc c = Just $! seq p' (if c == '\n' then seq l' (p', l', p) else (p', l, lp))
+            fc' (p, l, lp) c | fc c = Just $! seq p'
+                                      $ if c == '\n' || c == '\f' then seq l' (p', l', p)
+                                        else if c == '\r' then (p', l, p)
+                                        else if c == '\t' then (p', l, lp + (p - lp) `mod` 8 - 8)
+                                        else if isZeroWidth c then (p, l, lp)
+                                        else (p', l, lp)
                              | otherwise = Nothing
                where p' = succ p
                      l' = succ l
@@ -575,6 +657,8 @@ instance (StableFactorial m, TextualMonoid m) => TextualMonoid (LinePositioned m
    break_ bt fc t = span_ (not bt) (not . fc) t
    dropWhile_ bt fc t = snd (span_ bt fc t)
    takeWhile_ bt fc (LinePositioned p l lp t) = LinePositioned p l lp (takeWhile_ bt fc t)
+   toString ft lpt = toString (ft . pure) (extractLines lpt)
+   toText ft lpt = toText (ft . pure) (extractLines lpt)
 
    {-# INLINE characterPrefix #-}
    {-# INLINE splitCharacterPrefix #-}
@@ -597,16 +681,31 @@ instance (StableFactorial m, TextualMonoid m) => TextualMonoid (LinePositioned m
    {-# INLINE takeWhile_ #-}
 
 linesColumns :: TextualMonoid m => m -> (Int, Int)
-linesColumns t = Textual.foldl (const . fmap succ) fc (0, 0) t
-   where fc (l, _) '\n' = (succ l, 0)
+linesColumns t = Textual.foldl (const . fmap succ) fc (0, 1) t
+   where fc (l, _) '\n' = (succ l, 1)
+         fc (l, _) '\f' = (succ l, 1)
+         fc (l, _) '\r' = (l, 1)
+         fc (l, c) '\t' = (l, c + 9 - c `mod` 8)
+         fc (l, c) ch | isZeroWidth ch = (l, c)
          fc (l, c) _ = (l, succ c)
 linesColumns' :: TextualMonoid m => m -> (Int, Int)
-linesColumns' t = Textual.foldl' (const . fmap succ) fc (0, 0) t
-   where fc (l, _) '\n' = let l' = succ l in seq l' (l', 0)
+linesColumns' t = Textual.foldl' (const . fmap succ) fc (0, 1) t
+   where fc (l, _) '\n' = let l' = succ l in seq l' (l', 1)
+         fc (l, _) '\f' = let l' = succ l in seq l' (l', 1)
+         fc (l, _) '\r' = (l, 1)
+         fc (l, c) '\t' = (l, c + 9 - c `mod` 8)
+         fc (l, c) ch | isZeroWidth ch = (l, c)
          fc (l, c) _ = let c' = succ c in seq c' (l, c')
 {-# INLINE linesColumns #-}
 {-# INLINE linesColumns' #-}
 
+isZeroWidth :: Char -> Bool
+isZeroWidth '\x200b' = True  -- zero width space
+isZeroWidth '\x200c' = True  -- zero width non-joiner
+isZeroWidth '\x200d' = True  -- zero width joiner
+isZeroWidth '\xfeff' = True  -- zero width no-break space
+isZeroWidth _ = False
+
 const3 :: a -> b -> c -> d -> a
 const3 a _p _l _lp = a
 {-# INLINE const3 #-}
